/*
 * Random instance generator
 */

#include <algorithm>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int_distribution.hpp>
#include <iostream>
#include <fstream>
#include <vector>
#include "util/util.hpp"

#define MAX_SEED 10000
#define NUM_GENERATORS 5

using namespace std;



/**
 * Instance sizes
 */
int n[] = {20, 30};


/** Necessary data to define a job */
struct Job {
	int release;
	int deadline;
	int duedate;
	int process;
	int weight;
};

/** Comparators */
struct JobRPComparator {
	bool operator()(const Job& jobA, const Job& jobB) {
		return( (jobA.release + jobA.process) < (jobB.release + jobB.process) );
	}
};

/** Comparators */
struct JobReleaseComparator {
	bool operator()(const Job& jobA, const Job& jobB) {
		return( jobA.release < jobB.release + jobB.process );
	}
};


/** Random generator indices */
enum SeederIndex {
	Release = 0,
	Deadline,
	Duedate,
	Process,
	Weight
};


/**
 * Random generators
 */
vector<boost::random::mt19937> rnd_gen;
boost::random::mt19937 shuffler;


/*
 * Initialize seeds for random generator
 */
void initialize_seeds() {
	boost::random::mt19937 random_seeder;
	random_seeder.seed(1);

	boost::random::uniform_int_distribution<> dist(1, MAX_SEED);
	for( int i = 0; i < (int)rnd_gen.size(); ++i ) {
		rnd_gen[i].seed(dist(random_seeder));
	}
	shuffler.seed(dist(random_seeder));
}

/**
 * Randomly shuffler a job vector
 */
void random_shuffler(vector<Job> &jobs) {
	int n = jobs.size();
	Job tmp;
	for( int i = 0; i < n-1; ++i ) {
		boost::random::uniform_int_distribution<> dist(i, n-1);
		int k = dist(shuffler);
		if( k != i ) {
			tmp = jobs[i];
			jobs[i] = jobs[k];
			jobs[i] = tmp;
		}
	}
}


/**
 * Create job file
 */
void generate_job_file(char* filename, vector<Job> &jobs) {
	ofstream out(filename);
	out << jobs.size() << endl;
	for( int i = 0; i < (int)jobs.size(); ++i ) {
		for( int j = 0; j < (int)jobs.size(); ++j ) {
			out << "0 ";
		}
		out << endl;
	}
	for( int i = 0; i < (int)jobs.size(); ++i ) {
		out << jobs[i].release;
		out << " " << jobs[i].deadline;
		out << " " << jobs[i].process;
		out << " " << jobs[i].duedate;
		out << " " << jobs[i].weight;
		out << endl;
	}
	out.close();
}

/**
 * Create job file with setup times
 */
void generate_job_file(char* filename, vector<Job> &jobs, vector< vector<int> > &setup) {
	ofstream out(filename);
	out << jobs.size() << endl;
	for( int i = 0; i < (int)jobs.size(); ++i ) {
		for( int j = 0; j < (int)jobs.size(); ++j ) {
			out << setup[i][j] << " ";
		}
		out << endl;
	}
	for( int i = 0; i < (int)jobs.size(); ++i ) {
		out << jobs[i].release;
		out << " " << jobs[i].deadline;
		out << " " << jobs[i].process;
		out << " " << jobs[i].duedate;
		out << " " << jobs[i].weight;
		out << endl;
	}
	out.close();
}


/**
 * Problem set type 1, Chu'92
 */
void generate_problem_tardiness() {

	int num_n = sizeof(n)/sizeof(int);
	vector<Job> jobs;

	boost::random::uniform_int_distribution<> p(1, 10);
	boost::random::uniform_int_distribution<> w(1, 10);
	double pi1[] = {0, 0.5, 1.0, 1.5};
	double pi2[] = {0.05, 0.25, 0.5};

	char filename[256];

	int n_p1 = sizeof(pi1)/sizeof(double);
	int n_p2 = sizeof(pi2)/sizeof(double);

	for( int i = 0; i < num_n; ++i ) {
		for( int alpha = 0; alpha < n_p1; ++alpha ) {
			for( int beta = 0; beta < n_p2; ++beta ) {
				for( int k = 0; k < 5; ++k ) {
					jobs.clear();

					// generate processing times
					int sum_p = 0;
					for( int j = 0; j < n[i]; ++j ) {
						Job job;
						job.process = p(rnd_gen[Process]);
						job.weight = w(rnd_gen[Weight]);
						job.deadline = -1;
						jobs.push_back(job);

						sum_p += job.process;
					}
					int r_width = pi1[alpha]*sum_p;
					int tw_width = pi2[beta]*sum_p;

					boost::random::uniform_int_distribution<> r(0, r_width);
					boost::random::uniform_int_distribution<> d(0, tw_width);

					// generate release and due dates
					for( int j = 0; j < n[i]; ++j ) {
						jobs[j].release = r(rnd_gen[Release]);
						jobs[j].duedate = jobs[j].release + jobs[j].process + d(rnd_gen[Duedate]);
					}

					// generate instance with weights
					sprintf(filename, "./tardiness/tardiness-%d-%d-%d-w-%d.txt",
							n[i], (int)(pi1[alpha]*100.0), (int)(pi2[beta]*100.0), k);
					generate_job_file(filename, jobs);

					// generate instance without weights
					for( int j = 0; j < n[i]; ++j ) {
						jobs[j].weight = 1;
					}
					sprintf(filename, "./tardiness/tardiness-%d-%d-%d-nw-%d.txt",
							n[i], (int)(pi1[alpha]*100.0), (int)(pi2[beta]*100.0), k);
					generate_job_file(filename, jobs);
				}
			}
		}
	}
}


/**
 * Variable processing time, minimize makespan
 */
void generate_varproctime_makespan() {
	int num_n = sizeof(n)/sizeof(int);
	vector<Job> jobs;

	boost::random::uniform_int_distribution<> p(1, 100);

	double R[] = {0.2, 0.4, 0.6, 0.8, 1.0, 1.25, 1.5, 1.75, 2.0, 2.5, 3.0};
	int n_r = sizeof(R)/sizeof(double);

	char filename[256];

	for( int i = 0; i < num_n; ++i ) {
		for( int r = 0; r < n_r; ++r ) {
			for( int k = 0; k < 5; ++k ) {
				jobs.clear();

				// generate processing times
				for( int j = 0; j < n[i]; ++j ) {
					Job job;
					job.process = p(rnd_gen[Process]);
					job.weight = 1;
					job.deadline = -1;
					job.duedate = -1;
					jobs.push_back(job);
				}

				int range = ((double)n[i])*(50.5)*R[r];
				//cout << range << endl;

				boost::random::uniform_int_distribution<> rel(0, range);

				// generate release dates
				for( int j = 0; j < n[i]; ++j ) {
					jobs[j].release = rel(rnd_gen[Release]);
				}

				// generate instance with weights
				sprintf(filename, "./varproc-makespan/varproc-%d-%d-w-%d.txt", n[i], (int)(100.0*R[r]), k);
				generate_job_file(filename, jobs);
			}
		}
	}
}


/**
 * Variable processing time, minimize max tardiness
 */
void generate_varproctime() {

	int num_n = sizeof(n)/sizeof(int);
	vector<Job> jobs;

	boost::random::uniform_int_distribution<> p(1, 100);
	double pi1[] = {0.1, 0.2, 0.3, 0.4, 0.5};

	char filename[256];
	int n_p1 = sizeof(pi1)/sizeof(double);

	for( int i = 0; i < num_n; ++i ) {
		for( int alpha = 0; alpha < n_p1; ++alpha ) {
			for( int k = 0; k < 10; ++k ) {
				jobs.clear();

				// generate processing times
				int sum_p = 0;
				for( int j = 0; j < n[i]; ++j ) {
					Job job;
					job.process = p(rnd_gen[Process]);
					job.weight = 1;
					job.deadline = -1;
					job.release = 0;
					jobs.push_back(job);

					sum_p += job.process;
				}

				int min_due = (double)sum_p * (0.50 - pi1[alpha]);
				int max_due = (double)sum_p * (0.50 + pi1[alpha]);

				boost::random::uniform_int_distribution<> duedate(min_due, max_due);

				// generate due dates
				for( int j = 0; j < n[i]; ++j ) {
					jobs[j].duedate = duedate(rnd_gen[Duedate]);
				}

				// generate instance
				sprintf(filename, "./varproc/varproc-%d-%d-%d.txt",
						n[i], (int)(pi1[alpha]*100.0), k);
				generate_job_file(filename, jobs);
			}
		}
	}
}


/**
 * Problem set type 2, Potts'83
 */
void generate_problem_latency() {

	int num_n = sizeof(n)/sizeof(int);
	vector<Job> jobs;

	boost::random::uniform_int_distribution<> p(1, 100);
	boost::random::uniform_int_distribution<> w(1, 10);

	double R[] = {0.2, 0.4, 0.6, 0.8, 1.0, 1.25, 1.5, 1.75, 2.0, 2.5, 3.0};
	int n_r = sizeof(R)/sizeof(double);

	char filename[256];

	for( int i = 0; i < num_n; ++i ) {
		for( int r = 0; r < n_r; ++r ) {
			for( int k = 0; k < 5; ++k ) {
				jobs.clear();

				// generate processing times
				for( int j = 0; j < n[i]; ++j ) {
					Job job;
					job.process = p(rnd_gen[Process]);
					job.weight = w(rnd_gen[Weight]);
					job.deadline = -1;
					job.duedate = -1;
					jobs.push_back(job);
				}

				int range = ((double)n[i])*(50.5)*R[r];
				//cout << range << endl;

				boost::random::uniform_int_distribution<> rel(0, range);

				// generate release dates
				for( int j = 0; j < n[i]; ++j ) {
					jobs[j].release = rel(rnd_gen[Release]);
				}

				// generate instance with weights
				sprintf(filename, "./latency/latency-%d-%d-w-%d.txt", n[i], (int)(100.0*R[r]), k);
				generate_job_file(filename, jobs);

				// generate instance without weights
				for( int j = 0; j < n[i]; ++j ) {
					jobs[j].weight = 1;
				}
				sprintf(filename, "./latency/latency-%d-%d-nw-%d.txt", n[i], (int)(100.0*R[r]), k);
				generate_job_file(filename, jobs);
			}
		}
	}
}


/**
 * Problem set type 3, Pan and Shi'05 Type-I (on IEEE Transactions on Automation Science and Engineering)
 */
void generate_problem_set3() {

	int num_n = sizeof(n)/sizeof(int);
	vector<Job> jobs;

	boost::random::uniform_int_distribution<> p(1, 100);
	boost::random::uniform_int_distribution<> w(1, 10);
	double pi1[] = {0.5, 1};
	double pi2[] = {1, 2, 4, 8, 16};

	char filename[256];

	int n_p1 = sizeof(pi1)/sizeof(double);
	int n_p2 = sizeof(pi2)/sizeof(double);

	for( int i = 0; i < num_n; ++i ) {
		for( int alpha = 0; alpha < n_p1; ++alpha ) {
			for( int beta = 0; beta < n_p2; ++beta ) {
				for( int k = 0; k < 10; ++k ) {
					jobs.clear();

					// generate processing times
					int sum_p = 0;
					for( int j = 0; j < n[i]; ++j ) {
						Job job;
						job.process = p(rnd_gen[Process]);
						job.weight = w(rnd_gen[Weight]);
						job.duedate = -1;
						jobs.push_back(job);

						sum_p += job.process;
					}

					// generate release dates
					int r_width = pi1[alpha]*sum_p;
					boost::random::uniform_int_distribution<> r(0, r_width);
					for( int j = 0; j < n[i]; ++j ) {
						jobs[j].release = r(rnd_gen[Release]);
					}

					// generate deadlines
					sort(jobs.begin(), jobs.end(), JobRPComparator());
					int end_time = 0;
					for( vector<Job>::iterator it = jobs.begin(); it != jobs.end(); ++it ) {
						end_time = MAX(end_time, it->release) + it->process;
						boost::random::uniform_int_distribution<> deadline(0, pi2[beta]*(end_time-it->release));
						it->deadline = end_time + deadline(rnd_gen[Deadline]);
					}

					// generate instance with weights
					sprintf(filename, "./set3/set3-%d-%d-%d-%d.txt",
							n[i], (int)(pi1[alpha]*100.0), (int)(pi2[beta]), k);
					generate_job_file(filename, jobs);
				}
			}
		}
	}
}

/**
 * Problem set type 4, Gelinas and Soumis'97
 */
void generate_problem_set4() {

	int num_n = sizeof(n)/sizeof(int);
	vector<Job> jobs;

	boost::random::uniform_int_distribution<> p(1, 50);
	int width[] = {50, 75, 100, 125, 150, 175, 200, 225, 250, 275, 300};
	int n_width = sizeof(width)/sizeof(int);

	char filename[256];

	for( int i = 0; i < num_n; ++i ) {
		for( int w = 0; w < n_width; ++w ) {
			for( int k = 0; k < 10; ++k ) {
				jobs.clear();

				boost::random::uniform_int_distribution<> dist(1, width[w]);

				// generate processing times
				for( int j = 0; j < n[i]; ++j ) {
					Job job;
					job.process = p(rnd_gen[Process]);
					job.weight = 1;
					job.duedate = -1;
					jobs.push_back(job);
				}

				// randomly shuffler the vector
				random_shuffler(jobs);
				int min_release = 100;
				int sum_p = 0;
				for( int j = 0; j < n[i]; ++j ) {
					sum_p += jobs[j].process;
					jobs[j].release = sum_p - dist(rnd_gen[Release]);
					jobs[j].deadline = sum_p + jobs[j].process + dist(rnd_gen[Deadline]);
					if( j == 0 ) {
						min_release = jobs[j].release;
					} else {
						min_release = MIN(min_release, jobs[j].release);
					}
				}

				// adjust negative release times
				if( min_release < 0 ) {
					for( int j = 0; j < n[i]; ++j ) {
						jobs[j].release -= min_release;
						jobs[j].deadline -= min_release;
					}
				}

				// generate instance with weights
				sprintf(filename, "./set4/set4-%d-%d-%d.txt", n[i], width[w], k);
				generate_job_file(filename, jobs);
			}
		}
	}
}

/**
 * Problem set type 5 - setup times + release
 */
void generate_problem_set5() {

	int n_jobs = 15;  // only 10 jobs for this class

	char filename[256];

	for( int m = 0; m < 1; ++m ) {

		// generate setup times
		boost::random::uniform_int_distribution<> setup_dist(1, 10);
		int sum_setup = 0;
		vector< vector<int> > setup(n_jobs, vector<int>(n_jobs, 0));
		for( int i = 0; i < n_jobs; ++i ) {
			for( int j = 0; j < n_jobs; ++j ) {
				if( i == j ) {
					continue;
				}
				setup[i][j] = setup_dist(rnd_gen[Process]);
				sum_setup += setup[i][j];
			}
		}
		double avg_process = (sum_setup / n_jobs) + 1;

		vector<double> alpha;
		for( double a = 0; a <= 5.0; a += 0.1 ) {
			alpha.push_back(a);
		}
		int n_alpha = alpha.size();

		vector<Job> jobs;
		for( int a = 0; a < n_alpha; ++a ) {
			for( int k = 0; k < 20; ++k ) {
				jobs.clear();

				int max_dist = (int)(avg_process * alpha[a]);
				boost::random::uniform_int_distribution<> r(0, max_dist);
				for( int i = 0; i < n_jobs; ++i ) {
					Job job;
					job.release = r(rnd_gen[Release]);
					job.deadline = -1;
					job.duedate = -1;
					job.process = 0;
					job.weight = 1;

					jobs.push_back(job);
				}
				sprintf(filename, "./set5/set5-%d-%d-%d-%d.txt", n_jobs, m, (int)(10.0*alpha[a]), k);
				generate_job_file(filename, jobs, setup);
			}
		}
	}
}

/**
 * Problem set type 6: Importance of setup times
 */
void generate_problem_var_setup() {

	char filename[256];

	int num_n = sizeof(n)/sizeof(int);
	vector<Job> jobs;

//	vector<double> beta;
//	for( double b = 0.3; b <= 5.1; b += 0.5 ) {
//		beta.push_back(b);
//	}
//	int n_beta = beta.size();

	double alpha[] = {0.25, 0.5, 0.75};
	int n_alpha = sizeof(alpha)/sizeof(double);

	double beta[] = {0, 0.5, 1, 1.50, 2, 2.5, 3, 3.5, 4, 4.5, 5};
	int n_beta = sizeof(beta)/sizeof(double);

	for( int i = 0; i < num_n; ++i ) {
		int n_jobs = n[i];
		for( int a = 0; a < n_alpha; ++a ) {
			for( int m = 0; m < 1; ++m ) {
				// generate process times
				boost::random::uniform_int_distribution<> p(1, 100);
				jobs.clear();
				int sum_p = 0;
				for( int j = 0; j < n_jobs; ++j ) {
					Job job;
					job.process = p(rnd_gen[Process]);
					job.release = 0;
					job.deadline = -1;
					job.duedate = -1;
					job.weight = 1;
					jobs.push_back(job);

					sum_p += job.process;
				}

				// generate release times
				int r_width = alpha[a]*((double)sum_p);
				boost::random::uniform_int_distribution<> r(0, r_width);
				for( int j = 0; j < n_jobs; ++j ) {
					jobs[j].release = r(rnd_gen[Release]);
				}
				//sort(jobs.begin(), jobs.end(), JobReleaseComparator());

				for( int b = 0; b < n_beta; ++b ) {
					for( int k = 0; k < 10; ++k ) {
						// generate setup times
						int setup_range = beta[b] * (50.5);
						boost::random::uniform_int_distribution<> st_dist(0, setup_range);
						vector< vector<int> > setup(n_jobs, vector<int>(n_jobs, 0));
						for( int i = 0; i < n_jobs; ++i ) {
							for( int j = 0; j < n_jobs; ++j ) {
								if( i == j ) {
									continue;
								}
								setup[i][j] = st_dist(rnd_gen[Process]);
							}
						}
						sprintf(filename, "./varsetup/varsetup-%d-%d-%d-%d-%d.txt", n_jobs, m,
								(int)(alpha[a]*100.0), (int)(beta[b]*100.0), k);
						generate_job_file(filename, jobs, setup);

//						if( beta[b] == 0 ) {
//							break;
//						}
					}
				}
			}
		}
	}
}


int main_generator() {
//int main() {

	// create random generators
	rnd_gen.resize(NUM_GENERATORS);
	initialize_seeds();

	//generate_varproctime();
//	generate_varproctime_makespan();
//generate_problem_tardiness();
generate_problem_latency();
//	generate_problem_set2();
//	generate_problem_set3();
//	generate_problem_set4();
//	generate_problem_set5();
	//generate_problem_var_setup();

	return 0;
}






