/*
 * ------------------------------------------------------------
 * Implementation: Reader functions for TSPTW instances
 * ------------------------------------------------------------
 */

#include <iostream>
#include <fstream>
#include "tsptw_instances.hpp"


using namespace std;


/**
 * Read AFG TSPTW file
 */
void read_afg_input(IloEnv env, char* filename, ActivityArray &acts, int** &setup) {
	ifstream input(filename);
	if( !input.is_open() ) {
		cerr << "Error: could not open file " << filename << endl;
		exit(1);
	}

	int num_acts;

	input >> num_acts;
	setup = new int*[num_acts];
	for( int i = 0; i < num_acts; i++ ) {
		setup[i] = new int[num_acts];
		for( int j = 0; j < num_acts; j++ ) {
			input >> setup[i][j];

			// adjust negative setups to depot or from terminal
			if( (j == 0 && setup[i][j] < 0) || (i == num_acts-1 && setup[i][j] < 0) ) {
				setup[i][j] = 0;
			}
		}
	}

	acts.resize(num_acts);
	int release, deadline, processing;
	cout << "Activities: " << endl;
	for( int i = 0; i < num_acts; i++ ) {
		input >> release;
		input >> deadline;
		input >> processing;

		acts[i] = new Activity(env, i, release, deadline, processing);

		cout << "\t" << acts[i]->id;
		cout << "\tr:" << acts[i]->release;
		cout << "\td:" << acts[i]->deadline;
		cout << "\tp:" << acts[i]->processing;
		cout << endl;
	}

	input.close();
}


/**
 * Read DUMAS TSPTW file
 */
void read_dumas_input(IloEnv env, char* filename, ActivityArray &acts, int** &setup) {
	ifstream input(filename);
	if( !input.is_open() ) {
		cerr << "Error: could not open file " << filename << endl;
		exit(1);
	}

	int num_acts;

	input >> num_acts;

	// initialize setup
	setup = new int*[num_acts+1];
	for( int i = 0; i < num_acts+1; i++ ) {
		setup[i] = new int[num_acts+1];
		for( int j = 0; j < num_acts+1; j++ ) {
			setup[i][j] = 0;
		}
	}

	// read activities
	for( int i = 0; i < num_acts; i++ ) {
		for( int j = 0; j < num_acts; j++ ) {
			input >> setup[i][j];
		}
	}

	// fix last activity
	for( int i = 0; i < num_acts; i++ ) {
		setup[i][num_acts] = setup[i][0];
		setup[num_acts][i] = setup[0][i];
	}
	acts.resize(num_acts+1);
	int release, deadline;
	for( int i = 0; i < num_acts; i++ ) {
		input >> release;
		input >> deadline;
		acts[i] = new Activity(env, i, release, deadline, 0);;
	}

	// create last activity
	acts[num_acts] = new Activity(env, num_acts, acts[0]->release, acts[0]->deadline, 0);

	// fix counter
	num_acts++;

//	// adjust processing times
//	for( int i = 0; i < num_acts; ++i ) {
//	  int min_setup = INF;
//	  for( int j = 0; j < num_acts; ++j) {
//	    if( i != j ) {
//	      min_setup = MIN(min_setup, setup[i][j]);
//	    }
//	  }
//	  acts[i]->processing = min_setup;
//	  for( int j = 0; j < num_acts; ++j) {
//	    if( i != j ) {
//	      setup[i][j] -= min_setup;
//	    }
//	  }
//	}

	cout << "Activities: " << endl;
	for( int i = 0; i < num_acts; i++ ) {
		cout << "\t" << acts[i]->id;
		cout << "\tr:" << acts[i]->release;
		cout << "\td:" << acts[i]->deadline;
		cout << "\tp:" << acts[i]->processing;
		cout << endl;
	}

	input.close();
}


/**
 * Read Sequential Ordering Problems
 **/
void read_sop_input(IloEnv env, char* filename, ActivityArray &acts, int** &setup, bool** &precedes) {

	ifstream input(filename);
	if( !input.is_open() ) {
		cerr << "Error: could not open file " << filename << endl;
		exit(1);
	}

	char str_in[256];
	input >> str_in;
	while( strcmp(str_in, "EDGE_WEIGHT_SECTION") != 0 && strcmp(str_in, "edge_weight_section") != 0 ) {
		input >> str_in;
	}

	int num_acts;
	input >> num_acts;

	// initialize and read setup/precedence arrays
	setup = new int*[num_acts];
	precedes = new bool*[num_acts];
	for( int i = 0; i < num_acts; i++ ) {
		setup[i] = new int[num_acts];
		precedes[i] = new bool[num_acts];
		memset(precedes[i], false, sizeof(bool)*num_acts);
	}

	int sum_setup = 0;

	for( int i = 0; i < num_acts; i++ ) {
		for( int j = 0; j < num_acts; j++ ) {
			input >> setup[i][j];

			// if setup is negative, activity 'j' must precede activity 'i'
			if( setup[i][j] == -1 ) {
				precedes[j][i] = true;
				setup[i][j] = 0;
			} else {
				sum_setup += setup[i][j];
			}
		}
	}

	cout << "Sum of setup times: " << sum_setup << endl;

	// allocate activities
	acts.resize(num_acts);
	for( int i = 0; i < num_acts; i++ ) {
		acts[i] = new Activity(env, i, 0, 90000, 0, 0, 1);
	}
	input.close();

	cout << "Activities: " << endl;
	for( int i = 0; i < num_acts; i++ ) {
		cout << "\t" << acts[i]->id;
		cout << "\tr:" << acts[i]->release;
		cout << "\td:" << acts[i]->deadline;
		cout << "\tp:" << acts[i]->processing;
		cout << "\tdue:" << acts[i]->duedate;
		cout << "\tw:" << acts[i]->weight;
		cout << endl;
	}
}


/**
 * Read TSP with Pickup and Delivery instances
 **/
void read_tsppd_input(IloEnv env, char* filename, ActivityArray &acts, int** &setup, bool** &precedes) {

	ifstream input(filename);
	if( !input.is_open() ) {
		cerr << "Error: could not open file " << filename << endl;
		exit(1);
	}

	// num of activities
	int num_acts;
	input >> num_acts;

	// initialize and read setup/precedence arrays
	setup = new int*[num_acts+1];
	precedes = new bool*[num_acts+1];
	for( int i = 0; i < num_acts+1; i++ ) {
		setup[i] = new int[num_acts+1];
		precedes[i] = new bool[num_acts+1];

		memset(precedes[i], false, sizeof(bool)*(num_acts+1));
		memset(setup[i], 0, sizeof(int)*(num_acts+1));
	}

	// read coordinates and precedences
	int x,y;
	int index, kind, rel;
	vector< pair<int,int> > coords;

	// depot
	input >> index;
	input >> x;
	input >> y;
	coords.push_back( pair<int,int>(x,y) );

	// requests
	for( int i = 0; i < num_acts-1; ++i ) {
		input >> index;
		input >> x; input >> y;
		input >> kind; input >> rel;
		if( kind == 0 ) {
			// delivery
			precedes[index-1][rel-1] = true;
		}
		coords.push_back( pair<int,int>(x,y) );
	}

	// terminal
	coords.push_back( coords[0] );

	// distance matrix
	int sum_setup = 0;
	for( int i = 0; i < num_acts+1; i++ ) {
		for( int j = i+1; j < num_acts+1; j++ ) {
			setup[i][j] = (coords[i].first-coords[j].first)*(coords[i].first-coords[j].first)
					+ (coords[i].second-coords[j].second)*(coords[i].second-coords[j].second);
			setup[i][j] = round( sqrt( (double)setup[i][j] ) );
			setup[j][i] = setup[i][j];
			sum_setup += setup[i][j];
		}
	}

	cout << "Sum of setup times: " << sum_setup << endl;

	// allocate activities
	acts.resize(num_acts);
	for( int i = 0; i < num_acts; i++ ) {
		acts[i] = new Activity(env, i, 0, 90000, 0, 0, 1);
	}

	// add terminal
	acts.push_back( new Activity(env, num_acts, 0, 90000, 0, 0, 1) );
	input.close();

	cout << "Activities: " << endl;
	for( int i = 0; i < num_acts; i++ ) {
		cout << "\t" << acts[i]->id;
		cout << "\tr:" << acts[i]->release;
		cout << "\td:" << acts[i]->deadline;
		cout << "\tp:" << acts[i]->processing;
		cout << "\tdue:" << acts[i]->duedate;
		cout << "\tw:" << acts[i]->weight;
		cout << endl;
	}
}



/**
 * Read total tardiness instance
 */
void read_wt_input(IloEnv env, char* filename, ActivityArray &acts, int** &setup) {
    ifstream input(filename);
    if( !input.is_open() ) {
        cerr << "Error: could not open file " << filename << endl;
        exit(1);
    }

    int num_acts;
    input >> num_acts;

    // initialize setup
    setup = new int*[num_acts];
    for( int i = 0; i < num_acts; i++ ) {
        setup[i] = new int[num_acts];
        for( int j = 0; j < num_acts; j++ ) {
            setup[i][j] = 0;
        }
    }

    // read activity parametes
    acts.resize(num_acts);
    int processing, weight, duedate;
    for( int i = 0; i < num_acts; i++ ) {
        input >> processing;
        input >> weight;
        input >> duedate;
        acts[i] = new Activity(env, i, 0, INF, processing, duedate, weight);
    }

    cout << "Activities: " << endl;
    for( int i = 0; i < num_acts; i++ ) {
        cout << "\t" << acts[i]->id;
        //cout << "\tr:" << acts[i]->release;
        //cout << "\td:" << acts[i]->deadline;
        cout << "\tp:" << acts[i]->processing;
        cout << "\tdue:" << acts[i]->duedate;
        cout << "\tw:" << acts[i]->weight;
        cout << endl;
    }

    input.close();
}


/**
 * Read disjunctive instances
 */
void read_disj_input(IloEnv env, char* filename, ActivityArray &acts, int** &setup) {
  ifstream input(filename);
  if( !input.is_open() ) {
      cerr << "Error: could not open file " << filename << endl;
      exit(1);
  }

  int num_acts;
  input >> num_acts;

  // initialize setup
  setup = new int*[num_acts];
  for( int i = 0; i < num_acts; i++ ) {
      setup[i] = new int[num_acts];
      for( int j = 0; j < num_acts; j++ ) {
          setup[i][j] = 0;
      }
  }

  // read activities
  acts.resize(num_acts);
  int release, deadline, processing;
  for( int i = 0; i < num_acts; i++ ) {
    input >> processing;
    input >> release;
    input >> deadline;
    acts[i] = new Activity(env, i, release, deadline, processing);;
  }

  cout << "Activities: " << endl;
  for( int i = 0; i < num_acts; i++ ) {
      cout << "\t" << acts[i]->id;
      cout << "\tr:" << acts[i]->release;
      cout << "\td:" << acts[i]->deadline;
      cout << "\tp:" << acts[i]->processing;
      cout << endl;
  }

  input.close();
}


/**
 * Read randomly generated instances
 */
void read_random_input(IloEnv env, char* filename, ActivityArray &acts, int** &setup) {
  ifstream input(filename);
  if( !input.is_open() ) {
      cerr << "Error: could not open file " << filename << endl;
      exit(1);
  }

  int num_acts;
  input >> num_acts;

  // initialize setup
  setup = new int*[num_acts];
  for( int i = 0; i < num_acts; i++ ) {
      setup[i] = new int[num_acts];
      for( int j = 0; j < num_acts; j++ ) {
          input >> setup[i][j];
      }
  }

  // read activities
  acts.resize(num_acts);
  int release, deadline, processing, duedate, weight;
  for( int i = 0; i < num_acts; i++ ) {

    input >> release;
    input >> deadline;
	input >> processing;
	input >> duedate;
	input >> weight;
	if( deadline == -1 ) {
		deadline = INF;
	}
	if( duedate == -1 ) {
		duedate = INF;
	}
    acts[i] = new Activity(env, i, release, deadline, processing, duedate, weight);
  }

  cout << "Activities: " << endl;
  for( int i = 0; i < num_acts; i++ ) {
      cout << "\t" << acts[i]->id;
      cout << "\tr:" << acts[i]->release;
      cout << "\td:" << acts[i]->deadline;
      cout << "\tp:" << acts[i]->processing;
      cout << "\tdue:" << acts[i]->duedate;
      cout << "\tw:" << acts[i]->weight;
      cout << endl;
  }

  input.close();
}


/**
 * Read total tardiness instance with setup times
 */
void read_wts_input(IloEnv env, char* filename, ActivityArray &acts, int** &setup) {
    ifstream input(filename);
    if( !input.is_open() ) {
        cerr << "Error: could not open file " << filename << endl;
        exit(1);
    }

    char line[256];
    input.getline(line, 256); // instance name

    int num_acts;
    input.getline(line, 256, ':');
    input >> num_acts;

    // initialize setup
    setup = new int*[num_acts+1];
    for( int i = 0; i < num_acts+1; i++ ) {
        setup[i] = new int[num_acts+1];
        for( int j = 0; j < num_acts+1; j++ ) {
            setup[i][j] = 0;
        }
    }

    // initialize activities
    acts.resize(num_acts+1);
    for( int i = 0; i < num_acts+1; ++i ) {
      acts[i] = new Activity(env, i, 0, INF, 0, 0, 0);
    }

    for( int i = 0; i < 14; ++i ) {
      input.getline(line, 256); // generator parameters
    }

    // process times
    input.getline(line, 256);
    //cout << line << endl;
    int p;
    for( int i = 0; i < num_acts; ++i) {
      input >> p;
      acts[i+1]->set_processing(p);
    }
    input.getline(line, 256);

    // weights
    input.getline(line, 256);
    //cout << line << endl;
    for( int i = 0; i < num_acts; ++i) {
      input >> acts[i+1]->weight;
    }
    input.getline(line, 256);

    // due dates
    input.getline(line, 256);
    //cout << line << endl;
    for( int i = 0; i < num_acts; ++i) {
      input >> acts[i+1]->duedate;
    }
    input.getline(line, 256);

    // setup times
    input.getline(line, 256);
    int source, target, time;

    for( int i = 0; i < num_acts; ++i ) {
      input >> source;
      input >> target;
      input >> time;
      if( source != -1 ) {
        cerr << "Error reading WTS file: unexpected input" << endl;
        exit(1);
      }
      setup[0][target+1] = time;
    }
    for( int i = 0; i < num_acts; ++i ) {
      for( int j = 0; j < num_acts; ++j ) {
        if( i == j ) continue;
        input >> source;
        input >> target;
        input >> time;

        if( source != i || target != j ) {
          cerr << "Error reading WTS file: unexpected input" << endl;
          exit(1);
        }

        setup[i+1][j+1] = time;
      }
    }
    num_acts++;

    cout << "Activities: " << endl;
    for( int i = 0; i < num_acts; i++ ) {
        cout << "\t" << acts[i]->id;
        //cout << "\tr:" << acts[i]->release;
        //cout << "\td:" << acts[i]->deadline;
        cout << "\tp:" << acts[i]->processing;
        cout << "\tdue:" << acts[i]->duedate;
        cout << "\tw:" << acts[i]->weight;
        cout << endl;
    }

    input.close();
    //exit(1);
}

/**
 * Extract instance name
 */
void extract_instance_name(char* _filename, char* name) {
  char filename[256];
  strcpy(filename, _filename);
  char* c = strtok(filename, "/");
  while( c != NULL ) {
    strcpy(name, c);
    c = strtok (NULL, "/");
  }
  cout << name << endl;
}


