//============================================================================
// Main TSPTW file
// Read and solve a TSPTW instance
//============================================================================

#include <ilcp/cpext.h>
#include <iostream>
#include <vector>
#include <fstream>

#include "core/mdd_handler.hpp"
#include "core/search_goals.hpp"
#include "constraints/disjunctive.hpp"
#include "util/stats.hpp"
#include "util/util.hpp"

#include "instances/tsptw_instances.hpp"


// input types
#define DUMAS_INPUT   0     // dumas tsptw
#define AFG_INPUT     1     // afg tsptw
#define WT_INPUT      2     // weighted total tardiness
#define WTS_INPUT     3     // weighted total tardiness with setup times
#define DISJ_INPUT    4     // simple disjunctive
#define SOP_INPUT	   5    // sequential ordering problem
#define TSP_PD_INPUT  6     // TSP with pickup and delivery
#define RANDOM_INPUT  7     // randomly generated instances

// objective function types
#define FEASIBILITY 2

// search types
#define LEX_SEARCH 				0
#define DEFAULT_SEARCH 			1
#define MDD_SP_SEARCH 			2
#define MIN_SETUP_T_SEARCH 		3
#define NODEBRANCH_SEARCH 		4
#define DEFAULT_SEARCH_DFS		5


// parameters
#define TIME_LIMIT 86400


using namespace std;


/**
 * Simple constraint to impose precedences into a
 * sequence var object
 */
class ImposePrecedences : public IlcConstraintI {

public:

    /** Constructors */
	ImposePrecedences(IloCP cp,
			IlcIntervalVarArray _interval_vars,
			IlcIntervalSequenceVar _seq_var,
			bool** _precedes, int _num_tasks)
	: IlcConstraintI(cp), interval_vars(_interval_vars), seq_var(_seq_var),
	  precedes(_precedes), num_tasks(_num_tasks)
	{ }

    /** ILOG search control */
    virtual void propagate() {
		for( int i = 0; i < num_tasks; ++i ) {
			for( int j = 0; j < num_tasks; ++j ) {
				if( i != j && precedes[i][j] ) {
					//cout << "activity " << i << " precedes " << j << endl;
					seq_var.setBefore(interval_vars[i], interval_vars[j]);
				}
			}
		}
    }
    virtual void post() { }
    virtual IlcBool isViolated() const { return false;}

    /** Empty destructor */
    ~ImposePrecedences(){ };

private:
    IlcIntervalVarArray interval_vars;
    IlcIntervalSequenceVar seq_var;
    bool** precedes;
    int num_tasks;
};

/**
 * Wrapper 1
 */
inline IlcConstraint IlcImposePrecedences(
        IloCP cp,
        IlcIntervalVarArray _interval_vars,
        IlcIntervalSequenceVar _seq_var, bool** precedes, int num_tasks)
{
    return new(cp.getHeap()) ImposePrecedences(cp, _interval_vars, _seq_var, precedes, num_tasks);
}



/**
 * Wrapper for ILO object
 */
ILOCPCONSTRAINTWRAPPER4(IloImposePrecedences,
		cp,
		IloIntervalVarArray, _interval_vars,
		IloIntervalSequenceVar, _seq_var, bool**, precedes, int, num_tasks)
{
	// compose sequence var
	IlcIntervalSequenceVar seq_var;
	if( _seq_var.getImpl() != NULL ) {
		use(cp, _seq_var);
		seq_var = cp.getIntervalSequence(_seq_var);
	}

	// compose interval var
	IlcIntervalVarArray interval_vars(cp, _interval_vars.getSize());
	for( int i = 0; i < interval_vars.getSize(); i++ ) {
		use(cp, _interval_vars[i]);
		interval_vars[i] = cp.getInterval(_interval_vars[i]);
	}

	return IlcImposePrecedences(cp, interval_vars, seq_var, precedes, num_tasks);
}


/**
 * Solution
 */
struct Solution {
	vector<int> permutation;    // activity permutation
	double      time;           // time to reach solution
	int         branches;       // number of branches
	int         fails;          // number of fails
	int         makespan;       // makespan
	int         sum_setup;      // sum of setup times
	int         value;		    // solution objective function value
};


/**
 * Print solution to file
 */
void print_solution(
		ActivityArray &activities,
		IloCP& cp,
		IloIntVarArray& permutation,
		IloIntVar &makespan,
		IloIntVar &sum_setup,
		ofstream &out)
{
	// solution
	for( int i = 0; i < permutation.getSize(); i++ ) {
		out << cp.getValue(permutation[i]) << " ";
	}
	out << "\t[" << cp.getValue(makespan) << "]";
	out << "\t[" << cp.getValue(sum_setup) << "]";

	// search info
	out << "\t" << cp.getInfo(IloCP::NumberOfBranches);
	out << "\t" << cp.getInfo(IloCP::NumberOfFails);
	if( cp.getInfo(IloCP::FailStatus) == IloCP::SearchStoppedByLimit ) {
		out << "\tTL";
	} else {
		out << "\t" << cp.getTime();
	}

	out << endl;
}


/**
 * Main
 */
//int main(int argc, char* argv[]) {
int main_tsptw(int argc, char* argv[]) {

	if( argc < 6 || ( atoi(argv[5]) > 0 && argc < 7 ) ) {
		cout << "\nUsage: disjunctive [filename] [input-type] "
				"[obj-function] [search-type] [propagation-type] [width] "
				"[optional: refinement-type]\n" << endl;
		cout << "\twhere: " << endl;
		cout << "\t\tinput_type = 0: dumas instances" << endl;
		cout << "\t\tinput_type = 1: AFG instances" << endl;
		cout << "\t\tinput_type = 2: WT instances" << endl;
		cout << "\t\tinput_type = 3: WTS instances" << endl;
		cout << "\t\tinput_type = 4: disjunctive instances" << endl;
		cout << "\t\tinput_type = 5: sequential ordering problem" << endl;
		cout << "\t\tinput_type = 6: TSP with Pickup and Delivery" << endl;
		cout << "\t\tinput_type = 7: random instances" << endl;
		cout << endl;
		cout << "\t\tobj-function = 0: minimize makespan" << endl;
		cout << "\t\tobj-function = 1: minimize sum of setup times" << endl;
		cout << "\t\tobj-function = 2: minimize total tardiness" << endl;
		cout << "\t\tobj-function = 3: minimize max tardiness" << endl;
		cout << "\t\tobj-function = 4: minimize latency" << endl;
		cout << "\t\tobj-function = 5: none - only feasibility" << endl;
		cout << endl;
		cout << "\t\tsearch-type = 0: lexicographic search" << endl;
		cout << "\t\tsearch-type = 1: default search" << endl;
		cout << "\t\tsearch-type = 2: MDD shortest path search" << endl;
		cout << "\t\tsearch-type = 3: MDD min setup T search" << endl;
		cout << "\t\tsearch-type = 4: branch on MDD nodes" << endl;
		cout << "\t\tsearch-type = 5: default search with DFS" << endl;
		cout << endl;
		cout << "\t\tpropagation-type = 0: pure CP" << endl;
		cout << "\t\tpropagation-type = 1: pure MDD" << endl;
		cout << "\t\tpropagation-type = 2: CP+MDD" << endl;
		cout << endl;
		cout << "\t\trefinement-type = 0: deterministic according to objective (default)" << endl;
		cout << "\t\trefinement-type = 1: random" << endl;
		cout << "\t\trefinement-type = 2: read from file [order-file]" << endl;
		cout << endl << endl;

		exit(1);
	}


	try {

		IloEnv env;
		IloModel model(env);

		// ----------------------------------------------
		// Read input
		// ----------------------------------------------

		char inst_name[256];
		extract_instance_name(argv[1], inst_name);

		int input_type = atoi(argv[2]);
		DisjunctiveObjective obj_type = static_cast<DisjunctiveObjective>(atoi(argv[3]));
		int search_type = atoi(argv[4]);
		int propagation_type = atoi(argv[5]);

		int width = 0;

		ActivityArray acts;
		int** setup;
		bool** input_precedence;

		switch( input_type ) {
		case 0:
			read_dumas_input(env, argv[1], acts, setup);
			break;
		case 1:
			read_afg_input(env, argv[1], acts, setup);
			break;
		case WT_INPUT:
			read_wt_input(env, argv[1], acts, setup);
			break;
		case WTS_INPUT:
			read_wts_input(env, argv[1], acts, setup);
			break;
		case DISJ_INPUT:
			read_disj_input(env, argv[1], acts, setup);
			break;
		case SOP_INPUT:
			read_sop_input(env, argv[1], acts, setup, input_precedence);
			break;
		case TSP_PD_INPUT:
			read_tsppd_input(env, argv[1], acts, setup, input_precedence);
			break;
		case RANDOM_INPUT:
			read_random_input(env, argv[1], acts, setup);
			break;
		default:
			cout << "Error: instance type cannot be identified." << endl;
			exit(1);
		}

		cout << "\nInstance name: " << argv[1] << endl;

		// -----------------------------------------------
		// ILOG CPO Model
		// -----------------------------------------------

		/*
		 * Impose release/deadline constraint
		 */

		// set of ILO activities
		IloIntervalVarArray activities(env);

		// end times of activities
		IloIntExprArray end_times(env);

		for( ActivityArray::iterator act = acts.begin(); act != acts.end(); ++act ) {

			// release/deadline constraint
			model.add( IloStartOf((*act)->ilo_start_time) >= (*act)->release );
			model.add( IloEndOf((*act)->ilo_start_time) <= (*act)->deadline );

			end_times.add( IloEndOf((*act)->ilo_start_time) );

			// add to set of activities
			activities.add((*act)->ilo_start_time);
		}


		/*
		 * Permutation variables
		 */
		IloIntVarArray permutation;
		char varname[256];
		if( propagation_type == 1 || propagation_type == 2 ) {

			//permutation = IloIntVarArray(env, acts.size(), 0, acts.size()-1);

			permutation = IloIntVarArray(env, acts.size());
			for( int i = 0; i < (int)acts.size(); ++i ) {
				sprintf(varname, "permutation[%d]", i);
				//permutation[i] = IloIntVar(env, 0, acts.size()-1, varname);
				permutation[i] = IloIntVar(env, 0, acts.size()-1);
			}

			if( input_type == DUMAS_INPUT || input_type == AFG_INPUT || input_type == TSP_PD_INPUT ) {
				// first and last cities are depot
				permutation[0].setMax(0);
				permutation[acts.size()-1].setMin(acts.size()-1);
			} else if( input_type == WTS_INPUT ) {
				permutation[0].setMax(0);
			}

			// Enforce that all permutation variables must be distinct
			// Note: this is added outside the disjunctive constraint to allow the alldiff CPO propagator
			// to be performed first
			model.add( IloAllDiff(env, permutation) );
		}


		// objective function variable (if any)
		IloIntVar obj_var;
		if( obj_type != None ) {
			sprintf(varname, "obj_var[%d]", obj_type);
			obj_var = IloIntVar(env, 0, INF, varname);
			model.add( IloMinimize(env, obj_var) );
		}
		//model.add( obj_var <= 7500 );

		// create sequence variable with empty handle
		IloIntervalSequenceVar sequence;

		// ------------------------------------------------------
		// ILOG CP disjunctive propagator
		// ------------------------------------------------------
		if( propagation_type == 0 || propagation_type == 2 ) {

			// define activity times
			IloIntArray types(env, acts.size());
			for( unsigned int i = 0; i < acts.size(); i++ ) {
				types[i] = i;
			}

			// create sequence variables
			sequence = IloIntervalSequenceVar(env, activities, types);

			// TODO: change!! This just force extraction... (because of a deprecated function
			// in version 12.5
			for( unsigned int i = 0; i < acts.size(); i++ ) {
				model.add( IloTypeOfNext(sequence, acts[i]->ilo_start_time, -1) >= -100 );
			}


			if( input_type == DUMAS_INPUT || input_type == AFG_INPUT || input_type == TSP_PD_INPUT ) {
				// we fix first and last activity in the sequence
				model.add( IloFirst(env, sequence, acts[0]->ilo_start_time) );
				model.add( IloLast(env, sequence, acts.back()->ilo_start_time) );
			} else if( input_type == WTS_INPUT ) {
				model.add( IloFirst(env, sequence, acts[0]->ilo_start_time) );
			}

			// create transition cost matrix
			IloIntArray2   		  setup_costs(env, acts.size());
			IloTransitionDistance setup_times(env, acts.size());

			// Type of next for the last activity on the machine
			for( unsigned int i = 0; i < acts.size(); i++ ) {
				setup_costs[i] = IloIntArray(env, acts.size()+1);
				for( unsigned int j = 0; j < acts.size(); j++ ) {
					setup_times.setValue(i, j, setup[i][j]);
					setup_costs[i][j] = setup[i][j];

					if( input_type == SOP_INPUT ||  input_type == TSP_PD_INPUT ) {
						if( i != j && input_precedence[j][i] ) {
							setup_times.setValue(i, j, IloIntervalMax);
						}
					}
					///cout << i << "," << j << " --> " << setup_times.getValue(i,j) << endl;
				}
				setup_costs[i][acts.size()] = 0;
			}

			// check input precedences
			if( input_type == SOP_INPUT || input_type == TSP_PD_INPUT ) {
				//model.add( IloImposePrecedences(env, activities, sequence, input_precedence, acts.size()) );
				for( unsigned int i = 0; i < acts.size(); ++i ) {
					for( unsigned int j = i+1; j < acts.size(); ++j ) {

						if( input_precedence[i][j] ) {
							//cout << acts[i]->ilo_start_time << " << " << acts[j]->ilo_start_time << endl;

							model.add( IloBefore(env, sequence,
									acts[i]->ilo_start_time, acts[j]->ilo_start_time ) );
							model.add( IloEndBeforeStart(env,
									acts[i]->ilo_start_time,
									acts[j]->ilo_start_time) );

						} else if( input_precedence[j][i] ) {
							//cout << acts[j]->ilo_start_time << " << " << acts[i]->ilo_start_time << endl;

							model.add( IloBefore(env, sequence,
									acts[j]->ilo_start_time, acts[i]->ilo_start_time ) );
							model.add( IloEndBeforeStart(env,
									acts[j]->ilo_start_time,
									acts[i]->ilo_start_time) );
						}
					}
				}
			}

			// add CP optimizer disjunctive constraint
			model.add( IloNoOverlap(env, sequence, setup_times, IloTrue) );

			// create objective function
			switch( obj_type ) {

			case Makespan:
				model.add( obj_var == IloMax(end_times) );
				break;

			case SumSetupTimes: {

				if ( input_type == TSP_PD_INPUT ) {
					model.add( obj_var == IloMax(end_times) );
				} else {
					IloIntExpr obj(env);
					for( int i = 0; i < (int)acts.size(); i++ ) {
						obj += setup_costs[types[i]][IloTypeOfNext(sequence, acts[i]->ilo_start_time, acts.size())];
					}
					model.add( obj_var == obj);
				}
				break;
			}

			case TotalTardiness: {
				IloIntExpr tardiness_sum(env);
				for( int i = 0; i < (int)acts.size(); ++i ) {
					tardiness_sum += (acts[i]->weight) * IloMax(0, IloEndOf(acts[i]->ilo_start_time) - acts[i]->duedate);
				}
				model.add( obj_var == tardiness_sum );
				break;
			}

			case MaxTardiness: {

				// compute upper bound on max tardiness
				int max_release = 0;
				int max_weight = 0;
				int sum_p = 0;
				for( ActivityArray::iterator it = acts.begin(); it != acts.end(); ++it ) {
					max_release = MAX( max_release, (*it)->release );
					max_weight = MAX( max_weight, (*it)->weight );
					sum_p += (*it)->processing;
				}
				obj_var.setMax( (max_release + sum_p)*max_weight );

				IloIntExprArray tardiness_v(env);
				for( int i = 0; i < (int)acts.size(); ++i ) {
					tardiness_v.add((acts[i]->weight) * IloMax(0, IloEndOf(acts[i]->ilo_start_time) - acts[i]->duedate));
				}
				model.add( obj_var == IloMax( tardiness_v ) );
				break;
			}


			case Latency: {
				IloIntExpr latency_sum(env);
				for( int i = 0; i < (int)acts.size(); ++i ) {
					latency_sum += (acts[i]->weight) * IloEndOf(acts[i]->ilo_start_time);
				}
				model.add( obj_var == latency_sum );
				break;
			}


			default:
				break;
			}

			// add an an additional obj. function constraint for sequential ordering problem
			if( input_type == SOP_INPUT ) {
				model.add( obj_var == IloMax(end_times) );
			}
		}




//		// Link with next variables
//		if( propagation_type == 2 ) {
//
//			IloIntVarArray next_vars(env, acts.size()-1, 1, acts.size()-1);
//			for( int i = 0; i < (int)acts.size()-1; ++i ) {
//				model.add( next_vars[i] == IloTypeOfNext(sequence, activities[i], -1) );
//			}
//
//			for( int i = 0; i < (int)acts.size()-2; ++i ) {
//				model.add( next_vars[ permutation[i] ] == permutation[i+1] );
//			}
//
//		}



		// ------------------------------------------------------
		// MDD Propagator
		// ------------------------------------------------------

		DisjunctivePropagator* disj_prop = NULL;
		if( propagation_type == 1 || propagation_type == 2 ) {

			// read width
			width = atoi(argv[6]);
			if( width <= 0 ) {
				cout << "\nError: invalid width" << endl << endl;
				exit(1);
			}
			cout << "width: " << width << endl << endl;

			// create MDD disjunctive propagator

			IloArray<MDDPropagator*> propagators(env);

			if( obj_type != None ) {
				disj_prop = new DisjunctivePropagator(acts, sequence, permutation, setup,
						obj_type, obj_var);
			} else {
				disj_prop = new DisjunctivePropagator(acts, sequence, permutation, setup);
			}
			propagators.add( disj_prop );

			// create MDD handler with propagators and add it to model
			model.add( IloMDDHandler(env, permutation, activities, sequence, propagators, width, "mdd-handler") );


			// check input precedences
			if( input_type == SOP_INPUT || input_type == TSP_PD_INPUT ) {
				for( int i = 0; i < (int)acts.size(); ++i ) {
					for( int j = 0; j < (int)acts.size(); ++j ) {
						if( i == j ) continue;
						if( input_precedence[i][j] ) {
							disj_prop->set_precedence(i, j);
						}
					}
				}
			}

			// adjustment for some objective functions
			switch( obj_type ) {

			case Makespan:
				// TODO: check this!!
				model.add( obj_var == IloMax(end_times) );
				break;

			default:
				break;
			}

			// refinement type (if set)
			if( argc >= 8 ) {
				if( argv[7][0] == '1' ) {
					disj_prop->set_random_exact_refinement();
				} else if( argv[7][0] == '2' ) {
					vector<int> order( acts.size() );
					ifstream input_file( argv[8] );
					for( int i = 0; i < (int)acts.size(); ++i ) {
						input_file >> order[i];
					}
					input_file.close();
					disj_prop->set_exact_act_order( order );
				}
			}
		}


		// -----------------------------------------------
		// Solve
		// -----------------------------------------------

		IloCP cp(model);

		// set parameters
		cp.setParameter(IloCP::Workers, 1); // only one core
		cp.setParameter(IloCP::TimeLimit, TIME_LIMIT);
		cp.setParameter(IloCP::DefaultInferenceLevel, IloCP::Extended);
		cp.setParameter(IloCP::RelativeOptimalityTolerance, 1e-8);
//		cp.setParameter(IloCP::SearchType, IloCP::Restart);

		if( search_type != DEFAULT_SEARCH ) {
			cp.setParameter(IloCP::SearchType, IloCP::DepthFirst);
		}
		//cp.setParameter(IloCP::RestartFailLimit, 500);
		//cp.setParameter(IloCP::RestartProofEmphasis, 1000);
		//cp.setParameter(IloCP::LogVerbosity, IloCP::Quiet);

		//cp.setParameter(IloCP::RestartProofEmphasis, 1);

		//
		//		// generate all solutions
		//		cp.setParameter(IloCP::LogVerbosity, IloCP::Quiet);
		//		long int sol_count = 0;
		//		ofstream sols("sols.txt");
		//		cp.startNewSearch(search_procedure);
		//		while( cp.next() ) {
		//			if( propagation_type == 1 ) {
		//				for( int i = 0; i < permutation.getSize(); i++ ) {
		//					sols << cp.getValue(permutation[i]) << " ";
		//				}
		//				sols << endl;
		//			} else {
		//				IlcIntervalSequenceVar ilcseq = cp.getIntervalSequence(sequence);
		//				for( IloIntervalVar act = cp.getFirst(sequence); act.getImpl() != NULL; act = cp.getNext(sequence, act) ) {
		//				  sols << ilcseq.getType(cp.getInterval(act)) << " ";
		//				}
		//				//sols << " " << cp.getIntVar(makespan) << " " << cp.getIntVar(sum_setup);
		//				sols << endl;
		//			}
		//			sol_count++;
		//		}
		//		if( sol_count == 0 ) {
		//		  sols << "-1" << endl;
		//		}
		//		sols.close();
		//		exit(1);


		//		if( cp.propagate() ) {
		//			cout << cp.getIntVar(makespan) << endl;
		//			for( int i = 0; i < permutation.getSize(); i++ ) {
		//				cout << " " << cp.getIntVar(permutation[i]);
		//			}
		//		} else {
		//			cout << endl << endl << "Infeasible!" << endl << endl;
		//		}
		//		cout << endl;
		//		exit(1);

		//        if( cp.solve(search_phase) ) {


		// -----------------------------------------------
		// Search

//		if( cp.propagate() ) {
//			cout << "makespan: " << cp.getIntVar(obj_var) << endl;
//			if( permutation.getImpl() != NULL ) {
//				for( int i = 0; i < permutation.getSize(); ++i ) {
//					cout << "\t" << cp.getIntVar(permutation[i]) << endl;
//				}
//			}
//			for( int i = 0; i < acts.size(); ++i ) {
//				cout << "\t" << cp.getInterval(acts[i]->ilo_start_time) << endl;
//			}
//		}
//		exit(1);

		vector<Solution> all_solutions;   // all solutions found during search
		all_solutions.reserve(1000000);

		// set search strategy
		IloGoal search_procedure;
		bool is_goal = false;

		IloSearchPhase search_phase;
		bool is_phase = false;

		if( search_type == LEX_SEARCH ) {

			// create search method
			if( propagation_type == 1 ) {

				// Pure MDD: perform lexicographic search on permutation variables
				search_procedure = IloGenerate(env, permutation)
                    		  && IloSetTimesPermutation(env, permutation, activities);
				if( obj_type != None ) {
					search_procedure = search_procedure && IloFixIntVarMin(env, obj_var);
				}
				is_goal = true;

			} else {

				// CP or CP+MDD: lexicographic search on sequence of activities
				search_procedure = IloLexSchedule(env, activities, sequence)
                    		  && IloSetTimesSequence(env, sequence);

				if( obj_type != None ) {
					search_procedure = search_procedure && IloFixIntVarMin(env, obj_var);
				}
				is_goal = true;
			}
			//cp.startNewSearch(search_procedure);

		} else if( search_type == DEFAULT_SEARCH || search_type == DEFAULT_SEARCH_DFS ) {

			if( propagation_type == 1 ) {
				search_phase = IloSearchPhase(env, permutation);
			} else {
				IloIntervalSequenceVarArray seqs_array(env);
				seqs_array.add(sequence);
				search_phase = IloSearchPhase(env, seqs_array);
			}
			//cp.startNewSearch(search_phase);
			is_phase = true;

		} else if( search_type == MDD_SP_SEARCH ) {

			if( propagation_type == 0 ) {
				cout << "Error: Propagation type must include MDD" << endl;
				exit(1);
			}

			IloGoal search_procedure = IloShortestPathMDD(env, permutation, disj_prop)
                		&& IloSetTimesPermutation(env, permutation, activities);

			if( obj_type != None ) {
				search_procedure = search_procedure && IloFixIntVarMin(env, obj_var);
			}
			is_goal = true;

			//cp.startNewSearch(search_procedure);

		} else if( search_type == MIN_SETUP_T_SEARCH ) {

			if( propagation_type == 0 ) {
				cout << "Error: Propagation type must include MDD" << endl;
				exit(1);
			}

			search_procedure = IloSumSetupTimesMDD(env, permutation, disj_prop)
                		&& IloSetTimesPermutation(env, permutation, activities);

			if( obj_type != None ) {
				search_procedure = search_procedure && IloFixIntVarMin(env, obj_var);
			}
			is_goal = true;

			//cp.startNewSearch(search_procedure);


		} else if( search_type == NODEBRANCH_SEARCH ) {

			if( propagation_type == 0 || width <= 1 ) {
				cout << "Error: Propagation type must include MDD with witdh > 1" << endl;
				exit(1);
			}

			search_procedure = IloMDDNodeBranch(env, permutation, disj_prop)
                		&& IloSetTimesPermutation(env, permutation, activities);

			if( obj_type != None ) {
				search_procedure = search_procedure && IloFixIntVarMin(env, obj_var);
			}
			is_goal = true;

			//cp.startNewSearch(search_procedure);
		}

		bool feasible = false;
		if( is_goal ) {
			feasible = cp.solve( search_procedure );

		} else if( is_phase ) {
			feasible = cp.solve( search_phase );
		}

//		bool feasible = false;
//		while( cp.next() ) {
//
//			feasible = true;
//			if( obj_type == None ) {
//				break;
//			}
//
//			// store solution
//			Solution solution;
//			all_solutions.push_back(solution);
//			all_solutions.back().permutation.reserve(acts.size());
//
//			if( propagation_type != 0 ) {
//				for( int i = 0; i < permutation.getSize(); i++ ) {
//					all_solutions.back().permutation.push_back(cp.getValue(permutation[i]));
//				}
//			} else {
//				IlcIntervalSequenceVar ilcseq = cp.getIntervalSequence(sequence);
//				for( IloIntervalVar act = cp.getFirst(sequence); act.getImpl() != NULL; act = cp.getNext(sequence, act) ) {
//					all_solutions.back().permutation.push_back(ilcseq.getType(cp.getInterval(act)));
//				}
//			}
//			//          cout << "Makespan: " << cp.getIntVar(makespan) << endl;
//			//          cout << "Sum setup times: " << cp.getIntVar(sum_setup) << endl;
//			//          cout << "Total tardiness: " << cp.getIntVar(total_tardiness) << endl;
//			//          cout << endl;
//			all_solutions.back().value = cp.getValue(obj_var);
//			all_solutions.back().branches = cp.getInfo(IloCP::NumberOfBranches);
//			all_solutions.back().fails = cp.getInfo(IloCP::NumberOfFails);
//			all_solutions.back().time = cp.getTime();
//			//all_solutions.back().makespan = cp.getValue(makespan);
//		}

		int makespan = 0;
		int sum_setup = 0;

		if( feasible ) {
			if( propagation_type != 0 ) {
				cout << "\nSequence: " << endl;
				for( int i = 0; i < permutation.getSize(); i++ ) {
					//cout << cp.getValue(permutation[i]) << " ";
					cout << "\t";
					cout << "Activity " << cp.getValue(permutation[i]);
					cout << " :: " << cp.getInterval(activities[cp.getValue(permutation[i])]);
					cout << endl;

					makespan = MAX(makespan, cp.getInterval(activities[cp.getValue(permutation[i])]).getEndMax());
					if( i > 0 ) {
						sum_setup += setup[cp.getValue(permutation[i-1])][cp.getValue(permutation[i])];
					}
				}
				cout << endl;

			} else {
				cout << "\nSequence: " << endl;
				IlcIntervalSequenceVar ilcseq = cp.getIntervalSequence(sequence);
				int i = 0;
				int pos = 0, current, prev=-1;
				for( IloIntervalVar act = cp.getFirst(sequence); act.getImpl() != NULL; act = cp.getNext(sequence, act) ) {
					cout << "\t";
					cout << "Activity " << ilcseq.getType(cp.getInterval(act));
					if( permutation.getImpl() != NULL ) {
						cout << " (" << cp.getValue(permutation[i++]) << ")";
					}
					cout << " :: " << cp.getInterval(act);
					cout << endl;

					current = ilcseq.getType(cp.getInterval(act));
					makespan = MAX(makespan, cp.getInterval(activities[current]).getEndMax());
					if( pos > 0 ) {
						sum_setup += setup[prev][current];
					}
					pos++;
					prev = current;
				}
				cout << endl;
			}

			cout << "\nMakespan: " << makespan << endl;
			cout << "Sum setup times: " << sum_setup << endl;

			if( obj_type != None ) {
				cout << "\nBest solution: " << cp.getObjValue() << endl;
			} else {
				//cout << "\nMakespan: " << cp.getValue(makespan) << endl;
				//cout << "Sum of setup times: " << cp.getValue(sum_setup) << endl;
			}
		} else {
			cout << "\nProblem is infeasible." << endl;
		}
		cout << "\nBranches: " << cp.getInfo(IloCP::NumberOfBranches) << endl;
		cout << "Fails: " << cp.getInfo(IloCP::NumberOfFails) << endl;
		cout << "Time: " << cp.getTime() << endl << endl;

		// statistics
		char statname[256];
		sprintf(statname, "stats-%d-%d.txt", propagation_type, width);

		//ofstream statsfile("stats.txt", ios::app);
		ofstream statsfile(statname, ios::app);
		statsfile << argv[1];

		// objective function
		switch( obj_type ) {
		case Makespan:
			statsfile << "\tMakespan";
			break;

		case SumSetupTimes: {
			statsfile << "\tSumSetup";
			break;
		}

		case TotalTardiness: {
			statsfile << "\tTotalTardiness";
			break;
		}

		case MaxTardiness: {
			statsfile << "\tMaxTardiness";
			break;
		}

		case Latency: {
			statsfile << "\tLatency";
			break;
		}
		default:
			break;
		}


		switch( propagation_type ) {
		case 0:
			statsfile << "\tCP";
			break;
		case 1:
			statsfile << "\tMDD";
			break;
		case 2:
			statsfile << "\tCP+MDD";
			break;
		}

		if( propagation_type == 0 ) {
			statsfile << "\tnowidth";
		} else {
			statsfile << "\t" << width;
		}

		if( obj_type != None && feasible ) {
			statsfile << "\t[*" << cp.getValue(obj_var) << "]";
			statsfile << "\t[" << makespan << "]";
			statsfile << "\t[" << sum_setup << "]";
		} else if( feasible ) {
			statsfile << "\t[feasible]";
			statsfile << "\t[" << makespan << "]";
			statsfile << "\t[" << sum_setup << "]";
		} else {
			statsfile << "\t[infeasible]";
		}



		statsfile << "\t" << cp.getInfo(IloCP::NumberOfBranches);
		statsfile << "\t" << cp.getInfo(IloCP::NumberOfFails);

		//if( cp.getTime() >= TIME_LIMIT - 15) {
		if( cp.getInfo(IloCP::FailStatus) == IloCP::SearchStoppedByLimit ) {
			statsfile << "\tTL";
		} else {
			statsfile << "\t" << cp.getTime();
		}
		if( argc >= 8 ) {
			statsfile << "\t" << argv[7];
		}

		statsfile << endl;
		statsfile.close();

		// print all solutions
		char allsol_filename[256];
		sprintf(allsol_filename, "sols/sol_p%d_w%d_o%d_%s", propagation_type, width, obj_type, inst_name);

		ofstream allsols_file(allsol_filename);
		for( vector<Solution>::iterator sol = all_solutions.begin(); sol != all_solutions.end(); ++sol ) {
			for( int i = 0; i < (int)acts.size(); ++i ) {
				allsols_file << sol->permutation[i] << " ";
			}
			allsols_file << "\t[" << sol->value << "]";
			allsols_file << "\t" << sol->branches;
			allsols_file << "\t" << sol->fails;
			allsols_file << "\t" << sol->time;
			allsols_file << endl;
		}

	} catch (IloException &ex) {
		cout << "CPO Error: " << ex << endl;
	}

	return 0;
}
