//============================================================================
// Main Example
//============================================================================

#include <ilcp/cpext.h>
#include <iostream>
#include <vector>
#include <fstream>

#include "core/mdd_handler.hpp"
#include "core/search_goals.hpp"
#include "constraints/disjunctive.hpp"
#include "util/stats.hpp"
#include "util/util.hpp"

using namespace std;


/**
 * Main
 */
int main(int argc, char* argv[]) {

  if (argc <= 1) {
    cout << "\nUsage: mddCP [MDD width]" << endl << endl;
    exit(1);
  }

  try {
	  
    IloEnv env;
    IloModel model(env);

    DisjunctiveObjective obj_type = SumSetupTimes;

    // initialize activities
    int n_acts = 9;
    ActivityArray acts;
    for (int i = 0; i < n_acts; ++i) {
      acts.push_back( new Activity(env, i, 0, INF, 0) );
    }

    IloIntervalVarArray activities(env);
    for (int i = 0; i < n_acts; ++i) {
      activities.add( acts[i]->ilo_start_time );
    }
    
    // setup times
    int** setup = new int*[n_acts];
    for (int i = 0; i < n_acts; ++i) {
      setup[i] = new int[n_acts];
      for (int j = 0; j < n_acts; ++j) {
	setup[i][j] = i*n_acts + j;
      }
    }

    // create model
        
    char varname[256];

    /*
     * Permutation variables
     */
    IloIntVarArray permutation = IloIntVarArray(env, acts.size());
    for( int i = 0; i < (int)acts.size(); ++i ) {
      sprintf(varname, "permutation[%d]", i);
      permutation[i] = IloIntVar(env, 0, acts.size()-1);
    }

    // fix the first and last variable
    permutation[0].setMax(0);
    permutation[acts.size()-1].setMin(acts.size()-1);

    // Enforce that all variables must be pairwise different (redundant, but improves propagation)
    model.add( IloAllDiff(env, permutation) );

    // objective function variable (if any)
    IloIntVar obj_var;
    if( obj_type != None ) {
      sprintf(varname, "obj_var[%d]", obj_type);
      obj_var = IloIntVar(env, 0, INF, varname);
      model.add( IloMinimize(env, obj_var) );
    }

    // define activity types
    IloIntArray types(env, acts.size());
    for( unsigned int i = 0; i < acts.size(); i++ ) {
      types[i] = i;
    }

    // ---------------------------------------------------
    // No overlap - CP model (redundant)
    
    // create sequence variable (required for disjunctive constraint)
    IloIntervalSequenceVar sequence = IloIntervalSequenceVar(env, activities, types);

    // Type of next for the last activity on the machine
    IloTransitionDistance setup_times(env, acts.size());
    for (size_t  i = 0; i < acts.size(); ++i) {
      for (size_t j = 0; j < acts.size(); ++j) {
	setup_times.setValue(i, j, setup[i][j]);
      }
    }

    // add CP optimizer disjunctive constraint
    model.add( IloNoOverlap(env, sequence, setup_times, IloTrue) );

    // ---------------------------------------------------
    
    // read width
    int width = atoi(argv[1]);
    if( width <= 0 ) {
      cout << "\nError: invalid width" << endl << endl;
      exit(1);
    }
    cout << "width: " << width << endl << endl;


    // create MDD disjunctive propagator
    
    DisjunctivePropagator* disj_prop = new DisjunctivePropagator(
								 acts, 
								 sequence, 
								 permutation, 
								 setup,
								 obj_type, 
								 obj_var);

    IloArray<MDDPropagator*> propagators(env);    
    propagators.add( disj_prop );

    
    // create MDD handler with propagators and add it to model
    model.add( IloMDDHandler(env, permutation, activities, 
			     sequence, propagators, width, "mdd-handler") );


    // -----------------------------------------------
    // Solve
    // -----------------------------------------------

    IloCP cp(model);

    // set parameters
    cp.setParameter(IloCP::Workers, 1); // only one core
    cp.setParameter(IloCP::DefaultInferenceLevel, IloCP::Extended);
    cp.setParameter(IloCP::SearchType, IloCP::DepthFirst);

    // search phase
    IloIntervalSequenceVarArray sequence_array(env);
    sequence_array.add(sequence);
    IloSearchPhaseArray search_phases(env);
    search_phases.add(IloSearchPhase(env, sequence_array));

    if ( cp.solve(search_phases) ) {
      cout << "\nSequence: " << endl;
      for( int i = 0; i < permutation.getSize(); i++ ) {
	cout << "\t";
	cout << "Activity " << cp.getValue(permutation[i]);
	cout << " :: " << cp.getInterval(activities[cp.getValue(permutation[i])]);
	cout << endl;
      }
      cout << endl;
    } else {
      cout << "No solution was found." << endl;
    }
    
    
  } catch (IloException &ex) {
    cout << "CPO Error: " << ex << endl;
  }
    
  return 0;
}
